%% Implementation of the truncated Wirtinger Flow (TWF) algorithm proposed in the paper
%  ``Solving Random Quadratic Systems of Equations Is Nearly as Easy as Solving Linear Systems'' by Y. Chen and E. J. Cand?s.
%  The code below is adapted from implementation of the Wirtinger Flow algorithm designed and implemented by E. Candes, X. Li, and M. Soltanolkotabi

function Error = TWF(y, x, x0, Params, A, At)    

Params.grad_type   = 'TWF_Poiss'; 

Params.alpha_lb    = 0.3;
Params.mu          = 0.2;
Params.alpha_ub    = 5;
Params.alpha_h     = 5;
Params.alpha_y     = 3;

z = x0;

Error.MSE = zeros(Params.T,1);
Error.Alpha = zeros(Params.T,1);
Error.Sigma2 = zeros(Params.T,1);


%% Loop
grad_type = Params.grad_type;
if strcmp(grad_type, 'TWF_Poiss') == 1
    mu = @(t) Params.mu; % Schedule for step size 
elseif strcmp(grad_type, 'WF_Poiss') == 1
    tau0 = 330;                         % Time constant for step size
    mu = @(t) min(1-exp(-t/tau0), 0.2); % Schedule for step size  
end

z_old = zeros(length(x),1);
for t = 1: Params.T,
    grad = compute_grad(z, y, Params, A, At);
    z = z - mu(t) * grad;             % Gradient update 

    Error.MSE(t) = norm(x - exp(-1i*angle(trace(x'*z))) * z, 'fro')^2/norm(x,'fro')^2;
    Error.Alpha(t) = abs( x'*z/norm(x)^2 ) ;
    Error.Sigma2(t) = norm( z - x'*z/norm(x)^2 * x )^2/norm(x)^2; 
    
    %% stopping condition
    if norm(z-z_old)^2/norm(z)^2 < Params.stop_tol
        Error.MSE(t+1:Params.T) = Error.MSE(t);
        Error.Alpha(t+1:Params.T) = Error.Alpha(t);
        Error.Sigma2(t+1:Params.T) = Error.Sigma2(t);
        break;
    end
    z_old = z;
    
    if Params.show_hist
        fprintf('it = %d, alpha = %e, sigma2 = %e\n',t,Error.Alpha(t),Error.Sigma2(t));
    end
end


